# main.py
#
# Copyright 2024 Lo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
import gi

gi.require_version('Gtk', '4.0')
gi.require_version('Adw', '1')

from gi.repository import Gtk, Gio, Adw
from .window import NucleusWindow


class NucleusApplication(Adw.Application):
    def __init__(self):
        super().__init__(application_id='page.codeberg.lo_vely.Nucleus',
                         flags=Gio.ApplicationFlags.DEFAULT_FLAGS)
        self.create_action('quit', lambda *_: self.quit(), ['<primary>q'])
        self.create_action('about', self.on_about_action)
        self.create_action('search',
            lambda *_: self.props.active_window.searchbar.set_search_mode(True),
            ['<primary>f']
        )


    def do_activate(self):
        win = self.props.active_window
        if not win:
            win = NucleusWindow(application=self)
        win.present()


    def on_about_action(self, *args):
        about = Adw.AboutDialog.new_from_appdata(
            "/page/codeberg/lo_vely/Nucleus/page.codeberg.lo_vely.Nucleus.metainfo.xml",
            "3"
        )
        about.set_copyright("© 2024 Lo")
        # Translators: Replace "translator-credits" with your name/username, and optionally an email or URL.
        about.set_translator_credits(_('translator-credits'))
        about.add_legal_section(
            _("Data"),
            _("Chemistry data provided by Wikipedia"),
            Gtk.License.CUSTOM,
            _("Licensed under Attribution-ShareAlike 4.0 International \n<a href='https://creativecommons.org/licenses/by-sa/4.0/'>https://creativecommons.org/licenses/by-sa/4.0/</a>"),
        )
        about.present(self.props.active_window)


    def create_action(self, name, callback, shortcuts=None):
        action = Gio.SimpleAction.new(name, None)
        action.connect("activate", callback)
        self.add_action(action)
        if shortcuts:
            self.set_accels_for_action(f"app.{name}", shortcuts)


def main(version):
    app = NucleusApplication()
    return app.run(sys.argv)
    



# utils.py
#
# Copyright 2024 Lo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

def get_category_color(card, category: str, activatable: bool = True):
    color: str = ""

    if category == _('diatomic nonmetal') or category == _('polyatomic nonmetal'):
        color = "green"

    if category == _('alkali metal'):
        color = "red"
    elif category == _('transition metal'):
        color = "blue"
    elif category == _('noble gas'):
        color = "orange"
    elif category == _('metalloid'):
        color = "yellow"
    elif category == _('alkaline earth metal'):
        color = "purple"
    elif category == _('lanthanide'):
        color = "teal"
    elif category == _('post-transition metal'):
        color = "pink"
    elif category == _('actinide'):
        color = "slate"

    if not color == "":
        if activatable:
            card.add_css_class("suggested-action")
            card.add_css_class(color)
        else:
            card.add_css_class(f"elements-{color}")




# element_info.py
#
# Copyright 2024 Lo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import gi, json
import re as regex
from gi.repository import Adw, Gtk, Gio, GObject
from .utils import get_category_color


@Gtk.Template(resource_path="/page/codeberg/lo_vely/Nucleus/ui/property-card.ui")
class NucleusPropertyCard(Gtk.Box):
    __gtype_name__ = "NucleusPropertyCard"

    title = GObject.Property(type=str)
    value = GObject.Property(type=str)
    icon_name = GObject.Property(type=str)
    has_icon = GObject.Property(type=bool, default=True)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@Gtk.Template(resource_path="/page/codeberg/lo_vely/Nucleus/ui/property-row.ui")
class NucleusPropertyRow(Adw.ActionRow):
    __gtype_name__ = "NucleusPropertyRow"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@Gtk.Template(resource_path="/page/codeberg/lo_vely/Nucleus/ui/element-info.ui")
class NucleusElementInfo(Gtk.Box):
    __gtype_name__ = "NucleusElementInfo"

    element_card = Gtk.Template.Child()

    atomic_number = Gtk.Template.Child()
    element_symbol = Gtk.Template.Child()
    element_name = Gtk.Template.Child()
    element_property = Gtk.Template.Child()

    theoretical_indicator = Gtk.Template.Child()

    group_card = Gtk.Template.Child()
    period_card = Gtk.Template.Child()
    block_card = Gtk.Template.Child()

    protons_card = Gtk.Template.Child()
    electrons_card = Gtk.Template.Child()
    neutrons_card = Gtk.Template.Child()

    general_props = Gtk.Template.Child()
    physical_props = Gtk.Template.Child()
    atomic_props = Gtk.Template.Child()
    other_props = Gtk.Template.Child()

    electron_shell_row = Gtk.Template.Child()
    cpk_color_row = Gtk.Template.Child()
    cpk_color_preview = Gtk.Template.Child()

    data = None

    def __init__(self, data, **kwargs):
        super().__init__(**kwargs)

        self.data = data

        self.group_card.props.value = self.data['group']
        self.period_card.props.value = self.data['period']
        self.block_card.props.value = self.data['block']

        self.protons_card.props.value = self.data['number']
        self.electrons_card.props.value = self.data['number']
        self.neutrons_card.props.value = round(self.data['atomic_mass'] - self.data['number'])

        self.atomic_number.props.label = str(data['number'])
        self.element_symbol.props.label = data['symbol']
        self.element_name.props.label = data['name']
        self.element_property.props.label = data['category'].capitalize()
        get_category_color(self.element_card, data['category'], activatable=False)

        if "theoretical" in data:
            self.theoretical_indicator.props.visible = True

        if data['cpk-hex'] is not None:
            self.cpk_color_row.props.subtitle = f"#{data['cpk-hex']}"
            self.cpk_color_row.props.sensitive = True

            css_provider = Gtk.CssProvider()
            css_provider.load_from_string(f".cpk-color {{background: #{data['cpk-hex']};}}")
            self.cpk_color_preview.get_style_context().add_provider(
                css_provider,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
            )
            self.cpk_color_preview.add_css_class('cpk-color')
            self.cpk_color_preview.props.visible = True

        general_properties = [
            {'property_name':'summary', 'title':_("Summary")},
            {'property_name':'appearance', 'title':_("Appearance")},
            {'property_name':'atomic_mass', 'title':_("Atomic Mass"), 'unit':'u'},
        ]

        # The Kelvin sign used here is Unicode K (U+212A)
        physical_properties = [
            {'property_name':'phase', 'title':_("Phase"), 'icon':True},
            {'property_name':'density', 'title':_("Density"), 'unit':'g/cm<sup>3</sup>'},
            {'property_name':'melt', 'title':_("Melting Point"), 'unit':'K'},
            {'property_name':'boil', 'title':_("Boilling Point"), 'unit':'K'},
            {'property_name':'molar_heat', 'title':_("Molar Heat Capacity"), 'unit':'J/(mol·K)'}
        ]
        if "half-life" in self.data:
            unit = _("years") # Default unit
            if "half-life:unit" in self.data:
                unit = self.data["half-life:unit"]
            physical_properties.append({'property_name':'half-life', 'title': _("Half-life") + f" (<sup>{self.data['half-life:isotope']}</sup>{self.data['symbol']})", 'unit': unit})

        atomic_properties = [
            {'property_name':'electron_configuration_semantic', 'title':_("Electron Configuration")},
            {'property_name':'electron_affinity', 'title':_("Electron Affinity"), 'unit':'kJ/mol'},
            {'property_name':'electronegativity_pauling', 'title':_("Pauling Electronegativity")},
        ]

        other_properties = [
            {'property_name':'discovered_by', 'title':_("Discovered by")},
            {'property_name':'named_by', 'title':_("Named by")},
        ]

        for property in general_properties:
            self.add_property_row(property, self.general_props)

        for property in physical_properties:
            self.add_property_row(property, self.physical_props)

        for property in atomic_properties:
            self.add_property_row(property, self.atomic_props)

        for property in other_properties:
            self.add_property_row(property, self.other_props)


    def add_property_row(self, property, group):
        row = NucleusPropertyRow(
            title=property['title'],
            subtitle=f"{self.data[property['property_name']]}",
        )

        if (self.data[property['property_name']] == "" or self.data[property['property_name']] is None):
            row.props.subtitle = "-"
            row.props.sensitive = False
            group.add(row)
            return

        if 'unit' in property:
            if self.data['phase'] == 'Gas' and property['property_name'] == 'density':
                row.props.subtitle += " g/L"
            else:
                row.props.subtitle += f" {property['unit']}"

            if property['unit'] == 'K':
                row.props.subtitle = self.convert_temp(self.data[property['property_name']])

        if 'icon' in property:
            icon = Gtk.Image()

            if self.data['phase'] == _("Gas"):
                icon.props.icon_name = "gas-symbolic"
            elif self.data['phase'] == _("Solid"):
                icon.props.icon_name = "solid-symbolic"
            elif self.data['phase'] == _("Liquid"):
                icon.props.icon_name = "liquid-symbolic"

            row.add_suffix(icon)

        group.add(row)


    def convert_temp(self, kelvin) -> str:
        fahrenheit = ((kelvin - 273.15) * 1.8) + 32
        celcius = kelvin - 273.15

        return f"{kelvin} K = {fahrenheit:.2f} ℉ = {celcius:.2f} ℃"





# window.py
#
# Copyright 2024 Lo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import gi
import re as regex
from gi.repository import Adw, Gtk, Gio, GObject


from .grid_card import NucleusGridCard
from .element_info import NucleusElementInfo
from .electron_shell_dialog import ElectronShellDialog
from .utils import get_category_color


@Gtk.Template(resource_path="/page/codeberg/lo_vely/Nucleus/window.ui")
class NucleusWindow(Adw.ApplicationWindow):
    __gtype_name__ = "NucleusWindow"

    stack = Gtk.Template.Child()
    results_stack = Gtk.Template.Child()
    searchbar = Gtk.Template.Child()
    search_entry = Gtk.Template.Child()
    search_listbox = Gtk.Template.Child()
    results_count = 0

    bottom_sheet = Gtk.Template.Child()
    periodic_table = Gtk.Template.Child()
    split_view = Gtk.Template.Child()
    sidebar_scrolled_window = Gtk.Template.Child()
    close_sidebar_button = Gtk.Template.Child()
    legend = Gtk.Template.Child()

    last_selected_element: Gtk.ToggleButton = GObject.Property(type=Gtk.ToggleButton, default=None)
    current_element_name: str = GObject.Property(type=str, default=" ")
    show_sidebar_title: bool = GObject.Property(type=bool, default=False)
    element_source_link: str = GObject.Property(type=str)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.settings = Gio.Settings(schema_id="page.codeberg.lo_vely.Nucleus")
        self.settings.bind("width", self, "default-width", Gio.SettingsBindFlags.DEFAULT)
        self.settings.bind("height", self, "default-height", Gio.SettingsBindFlags.DEFAULT)
        self.settings.bind("is-maximized", self, "maximized", Gio.SettingsBindFlags.DEFAULT)

        from .periodic_table_data import data as json_file # Import translatable database

        for element in json_file["elements"]:
            row_symbol_center_box = Gtk.CenterBox(
                width_request=35,
                halign=Gtk.Align.CENTER,
            )
            row_symbol = Gtk.Label(
                label=element['symbol'],
                halign=Gtk.Align.CENTER,
                css_classes=["title-3"]
            )
            row = Adw.ActionRow(
                title=element['name'],
                subtitle=f"{element['number']} ⸱ {element['category'].title()}",
                activatable=True,
            )

            row_symbol_center_box.set_center_widget(row_symbol)
            row.add_prefix(row_symbol_center_box)
            row.element = element

            self.search_listbox.append(row)

            card = NucleusGridCard(element)
            get_category_color(card, element['category'])
            card.connect("clicked", self.on_grid_card_clicked, element)
            row.connect("activated", self.on_search_row_clicked, card, element)

            self.periodic_table.attach(
                card,
                element['xpos'],
                element['ypos'],
                width=1,
                height=1
            )

        #Grid numbers
        for number in range(1, 19):
            self.periodic_table.attach(
                child=Gtk.Label(
                    label=str(number),
                    margin_bottom=6,
                    css_classes=['dimmed']
                ),
                column=number,
                row=0,
                width=1,
                height=1
            )

        for number in range(1, 9):
            self.periodic_table.attach(
                child=Gtk.Label(
                    label=str(number),
                    margin_end=6,
                    css_classes=['dimmed']
                ),
                column=0,
                row=number,
                width=1,
                height=1
            )

        self.periodic_table.attach(self.legend, column=2, row=1, width=2, height=1)

        self.bottom_sheet.connect("notify::open", self.on_bottom_sheet_open_changed)

        self.search_listbox.set_filter_func(self.search_filter)
        self.searchbar.connect_entry(self.search_entry)
        self.searchbar.connect(
            "notify::search-mode-enabled",
            lambda *_: self.stack.set_visible_child_name (
                "search-view" if self.searchbar.get_search_mode() else "table-view"
            ),
        )

        self.search_entry.connect("search-changed", self.on_search_changed)
        self.sidebar_scrolled_window.props.vadjustment.connect("value-changed", self._on_vadjustment_value_changed)

        if self.props.default_width <= 445:
            self.searchbar.set_search_mode(True)


    def _on_vadjustment_value_changed(self, vadjustment) -> None:
        if vadjustment.props.value > 120:
            self.show_sidebar_title = True
        else:
            self.show_sidebar_title = False


    @Gtk.Template.Callback()
    def launch_source(self, _button) -> None:
        Gtk.UriLauncher(uri=self.props.element_source_link).launch(self, None)


    def search_filter(self, row):
        match = regex.search(
            self.search_entry.props.text,
            row.props.title + row.props.subtitle + row.element["symbol"],
            regex.IGNORECASE
        )
        if match:
            self.results_count += 1
        return match


    def on_search_changed(self, _search_widget):
        self.search_listbox.invalidate_filter()

        if self.results_count == 0:
            self.results_stack.set_visible_child_name("no-results")
        else:
            self.results_stack.set_visible_child_name("results")

        self.results_count = 0


    def on_search_row_clicked(self, row, grid_card, data):
        self.open_element_details(data)

        if self.last_selected_element:
            self.last_selected_element.props.active = False

        self.last_selected_element = grid_card
        self.last_selected_element.props.active = True


    def on_bottom_sheet_open_changed(self, _x, _y):
        if not self.bottom_sheet.props.open:
            self.close_element_details()


    @Gtk.Template.Callback()
    def close_element_details(self, _clicked_button=None):
        if self.last_selected_element is not None:
            self.last_selected_element.props.active = False
            self.last_selected_element.grab_focus()

        self.split_view.props.show_sidebar = False
        self.bottom_sheet.props.open = False


    def open_element_details(self, data):
        self.bottom_sheet.props.open = True
        self.split_view.props.show_sidebar = True
        self.load_chem_info(data)


    def on_grid_card_clicked(self, button, data):
        if self.last_selected_element is None:
            self.last_selected_element = button

        if button != self.last_selected_element:
            self.last_selected_element.props.active = False
            self.last_selected_element = button

            self.open_element_details(data)

        elif button == self.last_selected_element:
            if not self.split_view.props.show_sidebar:
                self.open_element_details(data)
            else:
                self.close_element_details()


    def load_chem_info(self, data) -> None:
        self.props.element_source_link = data['source']
        info = NucleusElementInfo(data)

        info.electron_shell_row.connect("activated", self._present_shell_dialog, data)
        self.current_element_name = data['name']

        if self.close_sidebar_button.props.visible:
            self.close_sidebar_button.grab_focus()

        self.sidebar_scrolled_window.props.child = info


    def _present_shell_dialog(self, _x, data) -> None:
        dialog = ElectronShellDialog(data['name'], data['shells'])
        dialog.present(self)



# grid_card.py
#
# Copyright 2024 Lo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import gi
from gi.repository import Gtk


@Gtk.Template(resource_path="/page/codeberg/lo_vely/Nucleus/ui/grid-card.ui")
class NucleusGridCard(Gtk.ToggleButton):
    __gtype_name__ = "NucleusGridCard"

    atomic_number = Gtk.Template.Child()
    symbol = Gtk.Template.Child()
    elem_property = Gtk.Template.Child()


    def __init__(self, data, **kwargs):
        super().__init__(**kwargs)

        self.atomic_number.props.label = str(data['number'])
        self.symbol.props.label = data['symbol']

        atomic_mass = f"{data['atomic_mass']:.3f}"
        self.elem_property.props.label = atomic_mass.rstrip('0').rstrip('.')




# electron_shell_dialog.py
#
# Copyright 2024 Lo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import gi, math

from gi.repository import Adw, Gtk, Gio, GObject


@Gtk.Template(resource_path="/page/codeberg/lo_vely/Nucleus/ui/electron-shell-dialog.ui")
class ElectronShellDialog(Adw.Dialog):
    __gtype_name__ = "ElectronShellDialog"

    drawing_area = Gtk.Template.Child()
    electron_pet_shell_row = Gtk.Template.Child()
    element_name = GObject.Property(type=str)

    drawing_color_light = [0,0,0,1]
    drawing_color_dark = [1,1,1,1]
    current_drawing_color = drawing_color_light

    shells = None

    def __init__(self, element_name, shells, **kwargs):
        super().__init__(**kwargs)

        self.props.element_name = element_name
        self.shells = shells

        shell_letter = ['K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R']
        for idx, shell in enumerate(self.shells):
            self.electron_pet_shell_row.props.subtitle += f"{shell_letter[idx]}{str(shell)} "

        self.drawing_area.set_content_height(len(self.shells) * 40 + 45)

        style_manager = Adw.StyleManager().get_default()
        style_manager.connect("notify::dark", self._on_color_scheme_changed)

        if style_manager.props.dark:
            self.current_drawing_color = self.drawing_color_dark

        self.drawing_area.set_draw_func(self._draw)


    def _on_color_scheme_changed(self, manager, dark):
        if manager.props.dark:
            self.current_drawing_color = self.drawing_color_dark
        else:
            self.current_drawing_color = self.drawing_color_light

        self.drawing_area.queue_draw()


    def _draw(self, _self, cr, width, height) -> None:
        cr.translate(width / 2, height / 2)

        for idx, shell in enumerate(self.shells, start=1):
            # Orbit
            edge_pos = idx * 20
            cr.set_source_rgba(
                self.current_drawing_color[0],
                self.current_drawing_color[1],
                self.current_drawing_color[2],
                self.current_drawing_color[3] - 0.5
            )
            cr.arc(0, 0, edge_pos, math.radians(0), math.radians(360))
            cr.stroke()

            for electron in range(shell):
                cr.set_source_rgba(
                    self.current_drawing_color[0],
                    self.current_drawing_color[1],
                    self.current_drawing_color[2],
                    self.current_drawing_color[3]
                )
                cr.arc(0, -edge_pos, 5, 0, 100)
                cr.rotate(math.radians(360 / shell))
                cr.fill()
                cr.stroke()
        
